@@ -1,6 +1,10 @@
-use std::convert::TryInto;
-use std::str;
-use std::str::FromStr;
+use std::{
+    convert::TryInto,
+    error::Error,
+    fmt::{Display, Formatter},
+    str,
+    str::FromStr,
+};
 
 pub const LENGTH: usize = 11;
 
@@ -9,12 +13,27 @@ pub const LENGTH: usize = 11;
 pub struct TurkishId([u8; LENGTH]);
 
 /// Represents the parser error for a given Turkish citizenship ID number.
-#[derive(Debug, Eq, PartialEq, Hash, Copy, Clone)]
+#[derive(Debug, Eq, PartialEq)]
 pub enum TurkishIdError {
     InvalidLength,
+    InvalidDigit,
     InvalidChecksum,
 }
 
+type Err = TurkishIdError;
+
+impl Display for Err {
+    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
+        match self {
+            Self::InvalidLength => write!(f, "invalid length"),
+            Self::InvalidDigit => write!(f, "invalid digit"),
+            Self::InvalidChecksum => write!(f, "invalid checksum"),
+        }
+    }
+}
+
+impl Error for Err {}
+
 impl TurkishId {
     fn new(value: &str) -> Self {
         TurkishId(
@@ -49,32 +68,40 @@ impl TurkishId {
 /// assert!(!trid::is_valid("06558242278"));
 /// ```
 pub fn is_valid(value: &str) -> bool {
-    let digits = value.as_bytes();
-    if digits.len() != LENGTH {
-        return false;
+    validate(value).is_ok()
+}
+
+macro_rules! next {
+    ($iter:ident) => {
+        $iter.next().flatten().ok_or(Err::InvalidDigit)
+    };
+}
+
+/// Internal function to validate a given Turkish ID number.
+fn validate(s: &str) -> Result<(), Err> {
+    if s.len() != LENGTH {
+        return Err(Err::InvalidLength);
     }
 
-    let mut invalid = false;
-    let mut odd_sum = digit(digits[0], &mut invalid);
-    if invalid || odd_sum == 0 {
-        return false;
+    let mut digits = s
+        .chars()
+        .map(|c| c.to_digit(10).and_then(|u| i32::try_from(u).ok()));
+    let mut odd_sum = next!(digits)?;
+    if odd_sum == 0 {
+        return Err(Err::InvalidDigit);
+    }
+    let mut even_sum = 0i32;
+    for _ in 0..4 {
+        even_sum += next!(digits)?;
+        odd_sum += next!(digits)?;
     }
 
-    let mut even_sum = 0;
-    for i in (1..9).step_by(2) {
-        even_sum += digit(digits[i], &mut invalid);
-        odd_sum += digit(digits[i + 1], &mut invalid);
-    }
-
-    let first_checksum = digit(digits[9], &mut invalid);
-    let final_checksum = digit(digits[10], &mut invalid);
-    if invalid {
-        return false;
-    }
+    let first_checksum = next!(digits)?;
+    let final_checksum = next!(digits)?;
 
     let computed_final = (odd_sum + even_sum + first_checksum) % 10;
     if computed_final != final_checksum {
-        return false;
+        return Err(Err::InvalidChecksum);
     }
 
     let mut computed_first = ((odd_sum * 7) - even_sum) % 10;
@@ -82,21 +109,22 @@ pub fn is_valid(value: &str) -> bool {
         computed_first += 10;
     }
 
-    computed_first == first_checksum
-}
-
-fn digit(byte: u8, invalid: &mut bool) -> i32 {
-    let b = (byte as i32) - ('0' as i32);
-    if b > 9 {
-        *invalid |= true;
-        return -1;
+    if computed_first != first_checksum {
+        return Err(Err::InvalidChecksum);
     }
-    b
+
+    Ok(())
 }
 
-impl ToString for TurkishId {
-    fn to_string(&self) -> String {
-        str::from_utf8(&self.0).expect("Invalid value").to_string()
+impl Display for TurkishId {
+    fn fmt(&self, f: &mut Formatter) -> Result<(), std::fmt::Error> {
+        write!(f, "{}", str::from_utf8(&self.0).unwrap())
+    }
+}
+
+impl From<&[u8; LENGTH]> for TurkishId {
+    fn from(value: &[u8; LENGTH]) -> Self {
+        TurkishId(value.to_owned())
     }
 }
 
@@ -132,14 +160,7 @@ impl FromStr for TurkishId {
     /// assert_eq!(result, Err(trid::TurkishIdError::InvalidLength));
     /// ```
     fn from_str(s: &str) -> Result<Self, Self::Err> {
-        if s.len() != LENGTH {
-            return Err(TurkishIdError::InvalidLength);
-        }
-
-        if !is_valid(s) {
-            return Err(TurkishIdError::InvalidChecksum);
-        }
-
+        validate(s)?;
         Ok(Self::new(s))
     }
 }
@@ -163,35 +184,35 @@ mod tests {
         "64404737702",
     ];
 
-    const INVALID_NUMBERS: &[&str] = &[
-        "04948892948", // first digit zero
-        "14948892946", // last digit INVALID
-        "14948892937", // last second digit INVALID
+    const INVALID_NUMBERS: &[(&str, Err)] = &[
+        ("04948892948", Err::InvalidDigit),    // first digit zero
+        ("14948892946", Err::InvalidChecksum), // last checksum INVALID
+        ("14948892937", Err::InvalidChecksum), // first checksum INVALID
         // non numeric chars
-        "A4948892948",
-        "7B558242278",
-        "80C76431508",
-        "767D5508630",
-        "9079E350894",
-        "43473F24496",
-        "566733G2584",
-        "2926080H600",
-        "93212606I04",
-        "352014085J8",
-        "3520140853K",
-        // uneven length
-        "7",
-        "76",
-        "76558",
-        "765582",
-        "7655824",
-        "76558242",
-        "765582422",
-        "7655824227",
-        "765582422781",
+        ("A4948892948", Err::InvalidDigit),
+        ("7B558242278", Err::InvalidDigit),
+        ("80C76431508", Err::InvalidDigit),
+        ("767D5508630", Err::InvalidDigit),
+        ("9079E350894", Err::InvalidDigit),
+        ("43473F24496", Err::InvalidDigit),
+        ("566733G2584", Err::InvalidDigit),
+        ("2926080H600", Err::InvalidDigit),
+        ("93212606I04", Err::InvalidDigit),
+        ("352014085J8", Err::InvalidDigit),
+        ("3520140853K", Err::InvalidDigit),
         // spaces
-        " 765582422781",
-        "765582422781 ",
+        (" 7655824227", Err::InvalidDigit),
+        ("5582422781 ", Err::InvalidDigit),
+        // uneven length
+        ("7", Err::InvalidLength),
+        ("76", Err::InvalidLength),
+        ("76558", Err::InvalidLength),
+        ("765582", Err::InvalidLength),
+        ("7655824", Err::InvalidLength),
+        ("76558242", Err::InvalidLength),
+        ("765582422", Err::InvalidLength),
+        ("7655824227", Err::InvalidLength),
+        ("765582422781", Err::InvalidLength),
     ];
 
     #[test]
@@ -203,8 +224,15 @@ mod tests {
 
     #[test]
     fn is_valid_invalidnumbers_returns_false() {
-        for number in INVALID_NUMBERS {
+        for (number, _) in INVALID_NUMBERS {
             assert!(!is_valid(number));
         }
     }
+
+    #[test]
+    fn parse_invalidnumbers_returns_correct_error() {
+        for (number, error) in INVALID_NUMBERS {
+            assert_eq!(*error, number.parse::<TurkishId>().err().unwrap());
+        }
+    }
 }
